unit Doc;

interface

uses
  Contnrs, Classes, Forms, SysUtils, DocIntf, Menus, ActnList;

type
  TDocKind = (dkUnknown, dkDoc, dkRep, dkWin);

  TDocManager = class
  private
    CRL: TClassList;     // All Forms Class ref List
    CL: TStringList;     // All Forms Caption List;
    DocCL: TStringList;  // Doc Caption List;
    FRepCL: TStringList; // Report Caption List;
    FWinCL: TStringList; // Window Caption List;
    FMenuItem: TMenuItem;
    FActionList: TActionList; 
    function GetCaptions(const TypeId: Integer): String;
    function GetTypes(const Caption: String): Integer;
    function CreateForm(const TypeId :Integer) :TForm;
    procedure ProcessParams(const D :IDoc; const Params :array of const);
    procedure ActionListExecute(Sender: TObject);
  protected
  public
    property MenuItem: TMenuItem read FMenuItem write FMenuItem;
    property ActionList: TActionList read FActionList;
    property WinCL :TStringList read FWinCL;
    property RepCL :TStringList read FRepCL;
    property Captions[const TypeId :Integer] :String read GetCaptions;
    property Types[const Caption :String] :Integer read GetTypes;
    procedure CreateMenuItems(AMenuItem :TMenuItem = nil);
    function GetDocKind(const TypeId :Integer) :TDocKind;
    function ShowDoc(const TypeId :Integer; const Params :array of const) :TForm;
    function ProcessDocParams(const TypeId :Integer; const Params :array of const) :TForm;
    procedure RegisterFormClass(const AClass: TClass; const TypeId: Integer;
      Caption: String; const DocKind: TDocKind = dkWin);
    constructor Create;
    destructor Destroy; override;
  end;

  EDMException = class(Exception);

var
  DocManager :TDocManager;

implementation

uses
  Misc;

{ TDocManager }

procedure TDocManager.ActionListExecute(Sender: TObject);
begin
  ShowDoc((Sender as TComponent).Tag, []);
end;

constructor TDocManager.Create;
begin
  inherited;
  CRL := TClassList.Create;
  CL := TStringList.Create;
  DocCL := TStringList.Create;
  FRepCL := TStringList.Create;
  FWinCL := TStringList.Create;
end;

function TDocManager.CreateForm(const TypeId: Integer): TForm;
begin
  if TypeId >= CRL.Count then raise EDMException.Create('    ');
  if not Assigned(CRL[TypeId]) then raise EDMException.Create('     ');
  Result := TFormClass(CRL[TypeId]).Create(Application);
  if TypeId < CL.Count then
    if Result.Caption <> CL[TypeId] then Result.Caption := CL[TypeId];
end;

type
  THackActionList = class(TActionList);

procedure TDocManager.CreateMenuItems(AMenuItem :TMenuItem);
var
  HAL :THackActionList;
  A :TAction;
  MI :TMenuItem;
  N :Integer;

 procedure AddFromList(List :TStringList);
 var
   I :Integer;
 begin
   for I := 0 to Pred(List.Count) do begin
     A := TAction.Create(FActionList);
     A.Caption := List[I];
     A.Tag := DocManager.Types[A.Caption];
     A.OnExecute := ActionListExecute;
     HAL.AddAction(A);
     MI := TMenuItem.Create(FMenuItem.Parent);
     MI.Action := A;
     MI.Caption := Format('&%d %s', [N, MI.Caption]);
     Inc(N);
     FMenuItem.Add(MI);
   end;
 end;
begin
  if Assigned(AMenuItem) then FMenuItem := AMenuItem;
  if not Assigned(FMenuItem) then Exit;
  if Assigned(FActionList) then FreeAndNil(FActionList);
  FActionList := TActionList.Create(FMenuItem.Owner);
  while 0 < FMenuItem.Count do FMenuItem.Delete(0);
  N := 1;
  HAL := THackActionList(FActionList);
  AddFromList(WinCL);
  MI := TMenuItem.Create(FMenuItem.Parent);
  MI.Caption := '-';
  FMenuItem.Add(MI);
  AddFromList(RepCL);
end;

destructor TDocManager.Destroy;
begin
  FreeAndNil(CRL);
  FreeAndNil(CL);
  FreeAndNil(DocCL);
  FreeAndNil(FRepCL);
  FreeAndNil(FWinCL);
  inherited;
end;

function TDocManager.GetCaptions(const TypeId: Integer): String;
begin
  if (CL.Count < Succ(TypeId)) or (TypeId < 0) then Result := '' else Result := CL[TypeId];
end;

function TDocManager.GetDocKind(const TypeId: Integer): TDocKind;
var
  S :String;
begin
  S := CL[TypeId];
  if DocCL.IndexOf(S) > -1 then Result := dkDoc
  else if RepCL.IndexOf(S) > -1 then Result := dkRep
  else if WinCL.IndexOf(S) > -1 then Result := dkWin
  else Result := dkUnknown;
end;

function TDocManager.GetTypes(const Caption: String): Integer;
begin
  Result := CL.IndexOf(Caption);
end;

function TDocManager.ProcessDocParams(const TypeId: Integer;
  const Params: array of const): TForm;
var
  I :Integer;
  D :IDoc;
  F, F1 :TForm;
begin
  Result := nil;
  if not Assigned(CRL[TypeId]) then Exit;
  with Screen do
    for I := 0 to Pred(FormCount) do
      if Forms[I].ClassType = CRL[TypeId] then
        if Supports(Forms[I], IDoc, D) then begin
          Result := Forms[I];
          ProcessParams(D, Params);
          Exit;
        end;
  if not Assigned(Result) then begin
    F := Screen.ActiveForm; F1 := nil;
    Result := CreateForm(TypeId);
    try
      D := Result as IDoc;
      D.CanEsc := True;
      D.CanClose := True;
      D.DocInit;
      ProcessParams(D, Params);
      F1 := Screen.ActiveForm;
    finally
      if GetDocKind(TypeId) = dkDoc then begin
        Result.Close;
        Result := nil;
      end
      else begin
        if Assigned(F) then F.Show;
        if (F1 <> F) and (F1 <> Result) and Assigned(F1) then F1.Show;
      end;
    end;
  end;
end;

procedure TDocManager.ProcessParams(const D: IDoc;
  const Params: array of const);
begin
  if not Assigned(D) then Exit;
  D.ProcessParams(VarOpenArrayToVarArray(Params));
end;

procedure TDocManager.RegisterFormClass(const AClass: TClass; const TypeId: Integer;
  Caption: String; const DocKind: TDocKind);
var
  I, N :Integer;
const
  sCF = '%s[%d]';
begin
  if CRL.IndexOf(AClass) <> TypeId then begin
    if CRL.Count < Succ(TypeId) then CRL.Count := Succ(TypeId);
    CRL[TypeId] := AClass;
    if Caption = '' then Caption := AClass.ClassName;
    if CL.Count < Succ(TypeId) then
      for I := 0 to TypeId - CL.Count do CL.Add('');
    if CL.IndexOf(Caption) > -1 then begin
      N := 0;
      for I := 2 to 10 do
        if CL.IndexOf(Format(sCF, [Caption, I])) = -1 then begin
          N := I;
          Break;
        end;
      if N > 0 then Caption := Format(sCF, [Caption, N]);
    end;
    CL[TypeId] := Caption;
    case DocKind of
      dkDoc: if DocCL.IndexOf(Caption) = -1 then DocCL.Add(Caption);
      dkRep: if RepCL.IndexOf(Caption) = -1 then RepCL.Add(Caption);
      dkWin: if WinCL.IndexOf(Caption) = -1 then WinCL.Add(Caption);
    end;
  end;
end;

function TDocManager.ShowDoc(const TypeId: Integer; const Params: array of const): TForm;
var
  I :Integer;
  D :IDoc;
  B :Boolean;
begin
  Result := nil;
  if not Assigned(CRL[TypeId]) then Exit;
  with Screen do
    for I := 0 to Pred(FormCount) do
      if Forms[I].ClassType = CRL[TypeId] then
        if Supports(Forms[I], IDoc, D) then begin
          if D.ParamsIs(Params) then begin
            Result := Forms[I];
            Break;
          end;
        end else begin
          Result := Forms[I];
          Break;
        end;
  if not Assigned(Result) then begin
    Result := CreateForm(TypeId);
    if Supports(Result, IDoc, D) then begin
      D.SetParams(Params);
      if not D.ParamsIs([]) then begin
        D.CanEsc := True;
        D.CanClose := True;
      end;
      D.DocInit;
    end;
  end;
  B := Application.MainForm.Enabled;
  Application.MainForm.Enabled := True;
  Result.Show;
  Application.MainForm.Enabled := B;
  ProcessParams(Result as IDoc, Params);
end;

initialization
  DocManager := TDocManager.Create;
finalization
  FreeAndNil(DocManager);
end.
